# ===========================================================================
# TestAutoencoder01.py: Test relationship between PCA and autoencoder for
#                       very simple example ... 
# 
# Written by: Mark Austin                                      September 2021 
# ===========================================================================

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns; sns.set()
import random
from sklearn.decomposition import PCA

import sklearn
from sklearn import datasets
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn import decomposition

import scipy
import tensorflow as tf
from keras.models import Model, load_model
from keras.layers import Input, Dense, Layer, InputSpec
from keras.callbacks import ModelCheckpoint, TensorBoard
from keras import regularizers, activations, initializers, constraints, Sequential
from keras import backend as K
from keras.constraints import UnitNorm, Constraint

# ============================================================== 
# Define noisy line function ...
# ============================================================== 

def NoisyLineFunction (a,b,x):
    return a + b*x + 1.0*(random.random() - 1.0)

# ============================================================== 
# Function to print matrices ...
# ============================================================== 

def PrintMatrix(a):
    for row in a:
        for col in row:
            print("{:8.3f}".format(col), end=" ")
        print("")

    print("--- Matrix size: ", a.size )
    print("--- Matrix shape: ", a.shape )

# ============================================================== 
# Create dense autoencoder with tied weights ...
# ============================================================== 

class DenseTied(Layer):
    def __init__(self, units,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 tied_to=None,
                 **kwargs):
        self.tied_to = tied_to
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super().__init__(**kwargs)
        self.units = units
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.input_spec = InputSpec(min_ndim=2)
        self.supports_masking = True
                
    def build(self, input_shape):
        assert len(input_shape) >= 2
        input_dim = input_shape[-1]

        if self.tied_to is not None:
            self.kernel = K.transpose(self.tied_to.kernel)
            self._non_trainable_weights.append(self.kernel)
        else:
            self.kernel = self.add_weight(shape=(input_dim, self.units),
                                          initializer=self.kernel_initializer,
                                          name='kernel',
                                          regularizer=self.kernel_regularizer,
                                          constraint=self.kernel_constraint)
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.units,),
                                        initializer=self.bias_initializer,
                                        name='bias',
                                        regularizer=self.bias_regularizer,
                                        constraint=self.bias_constraint)
        else:
            self.bias = None
        self.input_spec = InputSpec(min_ndim=2, axes={-1: input_dim})
        self.built = True

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) >= 2
        output_shape = list(input_shape)
        output_shape[-1] = self.units
        return tuple(output_shape)

    def call(self, inputs):
        output = K.dot(inputs, self.kernel)
        if self.use_bias:
            output = K.bias_add(output, self.bias, data_format='channels_last')
        if self.activation is not None:
            output = self.activation(output)
        return output

# ============================================================== 
# Constraint: Weights orthogonality.
# ============================================================== 

class WeightsOrthogonalityConstraint (Constraint):
    def __init__(self, encoding_dim, weightage = 1.0, axis = 0):
        self.encoding_dim = encoding_dim
        self.weightage = weightage
        self.axis = axis
        
    def weights_orthogonality(self, w):
        if(self.axis==1):
            w = K.transpose(w)
        if(self.encoding_dim > 1):
            m = K.dot(K.transpose(w), w) - K.eye(self.encoding_dim)
            return self.weightage * K.sqrt(K.sum(K.square(m)))
        else:
            m = K.sum(w ** 2) - 1.
            return m

    def __call__(self, w):
        return self.weights_orthogonality(w)

# ============================================================== 
# Constraint: Uncorrelated Encoded features
# ============================================================== 

class UncorrelatedFeaturesConstraint (Constraint):

    def __init__(self, encoding_dim, weightage=1.0):
        self.encoding_dim = encoding_dim
        self.weightage = weightage

    def get_covariance(self, x):
        x_centered_list = []

        for i in range(self.encoding_dim):
            x_centered_list.append(x[:, i] - K.mean(x[:, i]))

        x_centered = tf.stack(x_centered_list)
        covariance = K.dot(x_centered, K.transpose(x_centered)) / tf.cast(x_centered.get_shape()[0], tf.float32)

        return covariance

    # Constraint penalty
    def uncorrelated_feature(self, x):
        if(self.encoding_dim <= 1):
            return 0.0
        else:
            output = K.sum(K.square(
                self.covariance - tf.math.multiply(self.covariance, K.eye(self.encoding_dim))))
            return output

    def __call__(self, x):
        self.covariance = self.get_covariance(x)
        return self.weightage * self.uncorrelated_feature(x)

# ============================================================== 
# Main method ...
# ============================================================== 

def main():
    print("--- Enter TestAutoencoder01.main() ... ");
    print("--- ============================================= ... ");

    print("--- ");
    print("--- ================================================================= ... ");
    print("--- PART 1: Create data along noisy line function: y = ax + b + noise ... ");
    print("--- ================================================================= ... ");
    print("--- ");
    print("--- Populate x and y lists with data ... ");

    random.seed(21345);
    xvalues = list( np.arange( -10.0, 10.1, 0.25 ) );
    yvalues = []
    for x in xvalues:
        yvalue = NoisyLineFunction(2, 0.5, x)
        yvalues.append( yvalue );

    print("--- Transform x and y lists into 2d array ... ");

    X = np.transpose( [ xvalues, yvalues ] )

    PrintMatrix(X)

    print("--- ");
    print("--- Data statistics ... ");
    print("--- ");

    print("---   mean xvalue: ", X[:,0].mean() );
    print("---   mean yvalue: ", X[:,1].mean() );

    print("--- ");
    print("--- Plot raw data ... ");
    print("--- ");
 
    plt.scatter(X[:,0], X[:,1]) 
    plt.xlim(-15.0, 15.0 );
    plt.ylim(-15.0, 15.0 );
    plt.title('Mean-Centered Noisy Line Function');
    plt.xlabel('x');
    plt.ylabel('y');
    plt.show()

    print("--- ");
    print("--- ===================================================== ... ");
    print("--- PART 2: Principal Component Analysis (two dimensions) ... ");
    print("--- ===================================================== ... ");
    print("--- ");

    pca = PCA(n_components=2)
    pca.fit(X)

    print("--- Print principal components ... ");

    print( pca.components_ );

    print("--- Print variance ... ");

    print(pca.explained_variance_)

    print("--- ");
    print("--- ===================================================== ... ");
    print("--- PART 3: Autoencoder Data Processing ... ");
    print("--- ===================================================== ... ");
    print("--- ");

    print("--- Split data into train and test ... ");

    X_train, X_test = train_test_split(X, test_size=0.5, random_state=123)

    print("--- Data processing ... ");

    scaler = MinMaxScaler()
    scaler.fit(X_train)

    X_train_scaled = scaler.transform(X_train)
    X_test_scaled  = scaler.transform(X_test)

    print("--- Print scaled training data ... ");

    PrintMatrix( X_train )

    print("--- Print scaled testing data ... ");

    PrintMatrix( X_test )

    print("--- ");
    print("--- ============================================= ... ");
    print("--- PART 4: Fit Single Layer Autoencoder          ... ");
    print("--- ============================================= ... ");
    print("--- ");

    nb_epoch     = 1000
    batch_size   = 16
    input_dim    = X_train.shape[1]  # num of predictor variables, 
    encoding_dim  = 1
    learning_rate = 1e-3

    print('--- Input dim \n')
    print (input_dim)

    print('--- Create encoder and decoder ... \n')

    encoder = Dense(encoding_dim, activation="linear", input_shape=(input_dim,), use_bias = True) 
    decoder = Dense(   input_dim, activation="linear", use_bias = True)

    autoencoder = Sequential()
    autoencoder.add(encoder)
    autoencoder.add(decoder)

    print("--- Compile model ...")

    autoencoder.compile(metrics=['accuracy'], loss='mean_squared_error', optimizer='adam')

    print("--- Model summary ...")

    autoencoder.summary()

    print("--- Train model to fit data ...")

    autoencoder.fit( X_train, X_train, epochs=nb_epoch,
                     batch_size=batch_size, shuffle=True, verbose=2)

    print("--- ");
    print("--- ======================================================== ...");
    print("--- PART 5: Retrieve and print layer weights and bias values ...");
    print("--- ======================================================== ...");
    print("--- ");
    print("--- Trained model predictions ...")
    print("--- ");

    PrintMatrix( autoencoder.predict( X_test ) )

    print("--- ");
    print("--- Rounded model predictions ...")
    print("--- ");

    PrintMatrix( autoencoder.predict( X_test.round() ) )

    print("--- ");
    print("--- ======================================================================= ... ");
    print("--- PART 6: Compare and Contrast Outputs (Encoder and Decoder not the same) ... ");
    print("--- ======================================================================= ... ");
    print("--- ");

    np.set_printoptions(formatter={'float': '{: 0.5f}'.format})

    w_encoder = np.round( autoencoder.layers[0].get_weights()[0], 2).T 
    w_decoder = np.round( autoencoder.layers[1].get_weights()[0], 2)   

    print("--- Encoder weights:", w_encoder)
    print("--- Decoder weights:", w_decoder)

    print("--- Encoder: bias values ...")

    print ( autoencoder.layers[0].get_weights()[1] );

    print("--- Decoder: bias values ...")

    print ( autoencoder.layers[1].get_weights()[1] );

    print("--- Unlike PCA weights, the weights on Encoder and Decoder are not orthogonal ...")

    w_pca = pca.components_

    PrintMatrix( w_pca )

    np.round( np.dot(w_pca, w_pca.T), 2 )
    np.round( np.dot(w_encoder, w_encoder.T), 2 )
    np.round( np.dot(w_decoder, w_decoder.T), 2 )

    print("--- Unlike PCA features, i.e. Principal Scores, the Encoded features are correlated ...")

    pca_features = pca.fit_transform(X)
    np.round( np.cov(pca_features.T), 2)

    encoder_layer    = Model(inputs=autoencoder.inputs, outputs=autoencoder.layers[0].output)
    encoded_features = np.array(encoder_layer.predict(X_train_scaled))

    print("--- Encoded feature covariance:", np.cov(encoded_features.T))

    # Compute and print unit norm ...

    print('--- PCA weights norm, \n', np.sum(w_pca ** 2, axis = 1))
    print('--- Encoder weights norm, \n', np.sum(w_encoder ** 2, axis = 1))
    print('--- Decoder weights norm, \n', np.sum(w_decoder ** 2, axis = 1))

    # Train Test Reconstruction Accuracy

    train_predictions = autoencoder.predict(X_train_scaled)
    print('Train reconstrunction error\n', sklearn.metrics.mean_squared_error(X_train_scaled, train_predictions))

    test_predictions = autoencoder.predict(X_test_scaled)
    print('Test reconstrunction error\n', sklearn.metrics.mean_squared_error(X_test_scaled, test_predictions))

    print("--- ");
    print("--- ================================================================ ... ");
    print("--- PART 7: Optimize Autoencoder using PCA principles (tied weights) ... ");
    print("--- ================================================================ ... ");
    print("--- ");

    nb_epoch = 100
    batch_size = 16
    input_dim = X_train_scaled.shape[1] #num of predictor variables, 
    encoding_dim = 1
    learning_rate = 1e-3

    encoder = Dense(encoding_dim, activation="linear", input_shape=(input_dim,), use_bias = True) 
    decoder = DenseTied(input_dim, activation="linear", tied_to=encoder, use_bias = True)

    autoencoder = Sequential()
    autoencoder.add(encoder)
    autoencoder.add(decoder)

    autoencoder.compile(metrics=['accuracy'], loss='mean_squared_error', optimizer='adam')
    autoencoder.summary()

    autoencoder.fit(X_train_scaled, X_train_scaled, epochs=3, batch_size=batch_size,
                shuffle=True, verbose=0)

    print("--- Print encoder and decoder weights ... ");

    w_encoder = np.round( np.transpose(autoencoder.layers[0].get_weights()[0]), 3)
    w_decoder = np.round( autoencoder.layers[1].get_weights()[0], 3)

    print("--- Encoder weights:", w_encoder)
    print("--- Decoder weights:", w_decoder)

    print("--- Print train and test predictions ... ");

    train_predictions = autoencoder.predict(X_train_scaled)
    print('Train reconstrunction error\n', sklearn.metrics.mean_squared_error(X_train_scaled, train_predictions))
    test_predictions = autoencoder.predict(X_test_scaled)
    print('Test reconstrunction error\n', sklearn.metrics.mean_squared_error(X_test_scaled, test_predictions))

    print("--- ");
    print("--- ================================================================ ... ");
    print("--- PART 8: Optimize Autoencoder (Bias = True for Decoder)           ... ");
    print("--- ================================================================ ... ");
    print("--- ");

    encoder = Dense(encoding_dim, activation="linear", input_shape=(input_dim,), use_bias = True) 
    decoder = DenseTied(input_dim, activation="linear", tied_to=encoder, use_bias = True)

    autoencoder = Sequential()
    autoencoder.add(encoder)
    autoencoder.add(decoder)

    autoencoder.compile(metrics=['accuracy'], loss='mean_squared_error', optimizer='adam')
    autoencoder.summary()

    autoencoder.fit(X_train_scaled, X_train_scaled, epochs=nb_epoch, batch_size=batch_size,
                shuffle=True, verbose=0)

    w_encoder = np.round(np.transpose(autoencoder.layers[0].get_weights()[0]), 2)
    w_decoder = np.round(autoencoder.layers[1].get_weights()[1], 2)

    print('Encoder weights\n', w_encoder)
    print('Decoder weights\n', w_decoder)

    b_encoder = np.round(np.transpose(autoencoder.layers[0].get_weights()[1]), 2)
    b_decoder = np.round(np.transpose(autoencoder.layers[1].get_weights()[0]), 2)

    print("--- Encoder bias:", b_encoder)
    print("--- Decoder bias:", b_decoder)

    train_predictions = autoencoder.predict(X_train_scaled)
    print("--- Train reconstrunction error\n", sklearn.metrics.mean_squared_error(X_train_scaled, train_predictions))
    test_predictions = autoencoder.predict(X_test_scaled)
    print("--- Test reconstrunction error\n", sklearn.metrics.mean_squared_error(X_test_scaled, test_predictions))

    print("--- ");
    print("--- ================================================================ ... ");
    print("--- PART 9: Optimize Autoencoder (Encoder Weight Orthogonality)      ... ");
    print("--- ================================================================ ... ");
    print("--- ");

    encoder = Dense(encoding_dim, activation="linear", input_shape=(input_dim,),
                    use_bias=True,
                    kernel_regularizer=WeightsOrthogonalityConstraint(encoding_dim, weightage=1., axis=0)) 
    decoder = Dense(input_dim, activation="linear", use_bias = True)

    autoencoder = Sequential()
    autoencoder.add(encoder)
    autoencoder.add(decoder)

    autoencoder.compile(metrics=['accuracy'], loss='mean_squared_error', optimizer='adam')
    autoencoder.summary()

    autoencoder.fit(X_train_scaled, X_train_scaled, epochs=nb_epoch,
                batch_size=batch_size,
                shuffle=True,
                verbose=0)

    w_encoder = autoencoder.layers[0].get_weights()[0]
    print('Encoder weights dot product\n', np.round(np.dot(w_encoder.T, w_encoder), 2))

    train_predictions = autoencoder.predict(X_train_scaled)
    print('Train reconstrunction error\n', sklearn.metrics.mean_squared_error(X_train_scaled, train_predictions))
    test_predictions = autoencoder.predict(X_test_scaled)
    print('Test reconstrunction error\n', sklearn.metrics.mean_squared_error(X_test_scaled, test_predictions))

    print("--- ");
    print("--- ======================================================================== ... ");
    print("--- PART 10: Optimize Autoencoder (Encoder and Decoder Weight Orthogonality) ... ");
    print("--- ======================================================================== ... ");
    print("--- ");

    encoder = Dense(encoding_dim, activation="linear", input_shape=(input_dim,),
                    use_bias=True,
                    kernel_regularizer=WeightsOrthogonalityConstraint(encoding_dim, weightage=1., axis=0)) 
    decoder = Dense(input_dim, activation="linear", use_bias = True,
                    kernel_regularizer=WeightsOrthogonalityConstraint(encoding_dim, weightage=1., axis=1))

    autoencoder = Sequential()
    autoencoder.add(encoder)
    autoencoder.add(decoder)

    autoencoder.compile(metrics=['accuracy'], loss='mean_squared_error', optimizer='adam')
    autoencoder.summary()

    autoencoder.fit(X_train_scaled, X_train_scaled, epochs=nb_epoch,
                    batch_size=batch_size, shuffle=True, verbose=0)

    w_encoder = autoencoder.layers[0].get_weights()[0]
    print("--- Encoder weights dot product: ", np.round(np.dot(w_encoder.T, w_encoder), 2))

    w_decoder = autoencoder.layers[1].get_weights()[0]
    print("--- Decoder weights dot product ", np.round(np.dot(w_decoder, w_decoder.T), 2))

    print("--- ");
    print("--- ======================================================================= ... ");
    print("--- PART 11: Optimize Autoencoder (Unit Norm constraint on encoding layer)   ... ");
    print("--- ======================================================================= ... ");
    print("--- ");

    encoder = Dense(encoding_dim, activation="linear", input_shape=(input_dim,),
                    use_bias = True, kernel_constraint=UnitNorm(axis=0)) 
    decoder = Dense(input_dim, activation="linear", use_bias = True)

    autoencoder = Sequential()
    autoencoder.add(encoder)
    autoencoder.add(decoder)

    autoencoder.compile(metrics=['accuracy'], loss='mean_squared_error', optimizer='adam')
    autoencoder.summary()

    autoencoder.fit(X_train_scaled, X_train_scaled, epochs=nb_epoch, batch_size=batch_size,
                shuffle=True, verbose=0)

    w_encoder = np.round(autoencoder.layers[0].get_weights()[0], 2).T  # W in Figure 2.
    print('Encoder weights norm, \n', np.round(np.sum(w_encoder ** 2, axis = 1),2))

    # Training and test predictions ...

    train_predictions = autoencoder.predict(X_train_scaled)
    print('Train reconstrunction error\n', sklearn.metrics.mean_squared_error(X_train_scaled, train_predictions))
    test_predictions = autoencoder.predict(X_test_scaled)
    print('Test reconstrunction error\n', sklearn.metrics.mean_squared_error(X_test_scaled, test_predictions))

    print("--- ");
    print("--- ======================================================================= ... ");
    print("--- PART 12: Autoencoder (Unit Norm constraint on encoding/decoding layers) ... ");
    print("--- ======================================================================= ... ");
    print("--- ");

    encoder = Dense(encoding_dim, activation="linear", input_shape=(input_dim,),
                    use_bias = True, kernel_constraint=UnitNorm(axis=0)) 
    decoder = Dense(input_dim, activation="linear", use_bias = True, kernel_constraint=UnitNorm(axis=1))

    autoencoder = Sequential()
    autoencoder.add(encoder)
    autoencoder.add(decoder)

    autoencoder.compile(metrics=['accuracy'], loss='mean_squared_error', optimizer='adam')
    autoencoder.summary()

    autoencoder.fit(X_train_scaled, X_train_scaled, epochs=nb_epoch, batch_size=batch_size,
                shuffle=True, verbose=0)

    w_encoder = np.round(autoencoder.layers[0].get_weights()[0], 2).T  # W in Figure 2.
    w_decoder = np.round(autoencoder.layers[1].get_weights()[0], 2)  # W' in Figure 2.

    print('Encoder weights norm, \n', np.round(np.sum(w_encoder ** 2, axis = 1),2))
    print('Decoder weights norm, \n', np.round(np.sum(w_decoder ** 2, axis = 1),2))

    # Training and test predictions ...

    train_predictions = autoencoder.predict(X_train_scaled)
    print("--- Train reconstrunction error \n", sklearn.metrics.mean_squared_error(X_train_scaled, train_predictions))

    test_predictions = autoencoder.predict(X_test_scaled)
    print("--- Test reconstrunction error \n", sklearn.metrics.mean_squared_error(X_test_scaled, test_predictions))

    print("--- ");
    print("--- ======================================================================= ... ");
    print("--- PART 13: Autoencoder (put constraints together)                         ... ");
    print("--- ======================================================================= ... ");
    print("--- ");

    encoder = Dense(encoding_dim, activation="linear", input_shape=(input_dim,),
                    use_bias = True,
                    kernel_regularizer=WeightsOrthogonalityConstraint(encoding_dim, weightage=1., axis=0),
                    kernel_constraint=UnitNorm(axis=0)) 
    decoder = DenseTied(input_dim, activation="linear", tied_to=encoder, use_bias = False)

    autoencoder = Sequential()
    autoencoder.add(encoder)
    autoencoder.add(decoder)

    autoencoder.compile(metrics=['accuracy'], loss='mean_squared_error', optimizer='adam')
    autoencoder.summary()

    autoencoder.fit(X_train_scaled, X_train_scaled, epochs=nb_epoch, batch_size=batch_size, shuffle=True, verbose=0)

    # Training and test predictions ...

    train_predictions = autoencoder.predict(X_train_scaled)
    print('Train reconstrunction error\n', sklearn.metrics.mean_squared_error(X_train_scaled, train_predictions))
    test_predictions = autoencoder.predict(X_test_scaled)
    print('Test reconstrunction error\n', sklearn.metrics.mean_squared_error(X_test_scaled, test_predictions))

    print("--- ============================================= ... ");
    print("--- Leave TestAutoencoder01.main() ... ");

# call the main method ...

main()
